//  -*- Mode: C++ -*-

// inference-caches.cc

/*--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2003              |
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------*/

#include "logic/logic-system.hh"

namespace logic {
  using namespace stella;

boolean worldStateP(Context* self) {
  if (subtypeOfP(safePrimaryType(self), SGT_INFERENCE_CACHES_STELLA_WORLD)) {
    { Context* self000 = self;
      World* self = ((World*)(self000));

      return ((!((boolean)(((Context*)(dynamicSlotValue(self->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF, NULL)))))) &&
          (!((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_MONOTONICp, FALSE_WRAPPER)))->wrapperValue));
    }
  }
  else {
    return (TRUE);
  }
}

Context* getWorldState(Context* self) {
  return ((worldStateP(self) ? self : ((World*)(self))->baseModule));
}

World* lookupInferenceCache(Context* context, Keyword* kind) {
  { Context* cacheworld = NULL;

    if (kind == KWD_INFERENCE_CACHES_JUST_IN_TIME) {
      cacheworld = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL)));
    }
    else if (kind == KWD_INFERENCE_CACHES_META) {
      cacheworld = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, NULL)));
    }
    else if (kind == KWD_INFERENCE_CACHES_TMS) {
      cacheworld = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL)));
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << kind << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
    return (((World*)(cacheworld)));
  }
}

void initializeInferenceWorld(World* world) {
  setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_EVALUATION_STATE_TABLE, newHashTable(), NULL);
  setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_EVALUATION_QUEUE, newList(), NULL);
}

World* createInferenceCache(Context* context, Keyword* kind) {
  { World* world = createWorld((getQuotedTree("((:JUST-IN-TIME :TMS) \"/LOGIC\")", "/LOGIC")->memberP(kind) ? getInferenceCache(context, KWD_INFERENCE_CACHES_META) : context), NULL);

    if (kind == KWD_INFERENCE_CACHES_JUST_IN_TIME) {
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, world, NULL);
    }
    else if (kind == KWD_INFERENCE_CACHES_META) {
      setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_MONOTONICp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
      setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, ((!allPropositions(context->baseModule, TRUE)->nextP()) ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, world, NULL);
    }
    else if (kind == KWD_INFERENCE_CACHES_TMS) {
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, world, NULL);
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << kind << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
    setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF, context, NULL);
    initializeInferenceWorld(world);
    return (world);
  }
}

World* getInferenceCache(Context* context, Keyword* kind) {
  if (!worldStateP(context)) {
    return (((World*)(context)));
  }
  { World* world = lookupInferenceCache(context, kind);

    if (((boolean)(world))) {
      return (world);
    }
    processCheckTypesAgenda();
    world = createInferenceCache(context, kind);
    if (kind == KWD_INFERENCE_CACHES_JUST_IN_TIME) {
    }
    else if ((kind == KWD_INFERENCE_CACHES_META) ||
        (kind == KWD_INFERENCE_CACHES_TMS)) {
      { Module* parentmodule = NULL;
        Cons* iter000 = context->baseModule->parentModules->theConsList;

        while (!(iter000 == NIL)) {
          parentmodule = ((Module*)(iter000->value));
          iter000 = iter000->rest;
          inheritCachedInferences(lookupInferenceCache(parentmodule, kind), world);
        }
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << kind << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
    return (world);
  }
}

void destroyInferenceCache(Context* context, Keyword* kind) {
  { World* cache = NULL;

    if (kind == KWD_INFERENCE_CACHES_JUST_IN_TIME) {
      cache = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL)));
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL, NULL);
    }
    else if (kind == KWD_INFERENCE_CACHES_META) {
      cache = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, NULL)));
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, NULL, NULL);
    }
    else if (kind == KWD_INFERENCE_CACHES_TMS) {
      cache = ((World*)(dynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL)));
      setDynamicSlotValue(context->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL, NULL);
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << kind << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
    if (((boolean)(cache)) &&
        (!cache->deletedP())) {
      destroyWorld(cache);
    }
  }
}

void unlinkInferenceCache(World* self) {
  if (((boolean)(((Context*)(dynamicSlotValue(self->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF, NULL)))))) {
    { Context* worldstate = ((Context*)(dynamicSlotValue(self->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF, NULL)));

      if (self == ((World*)(dynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL)))) {
        setDynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL, NULL);
      }
      else if (self == ((World*)(dynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, NULL)))) {
        setDynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE, NULL, NULL);
      }
      else if (self == ((World*)(dynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL)))) {
        setDynamicSlotValue(worldstate->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL, NULL);
      }
      setDynamicSlotValue(self->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF, NULL, NULL);
    }
  }
}

void destroyInferenceCaches(Context* context) {
  destroyInferenceCache(context, KWD_INFERENCE_CACHES_JUST_IN_TIME);
  destroyInferenceCache(context, KWD_INFERENCE_CACHES_META);
  destroyInferenceCache(context, KWD_INFERENCE_CACHES_TMS);
}

// Used to signal an assertion that doesn't
// blow away inference caches.
DEFINE_STELLA_SPECIAL(oINVISIBLEASSERTIONpo, boolean , FALSE);

boolean metaPropositionP(Proposition* proposition) {
  if (!((boolean)(proposition))) {
    return (FALSE);
  }
  { boolean pointstometaobjectP = FALSE;

    { Object* arg = NULL;
      Vector* vector000 = proposition->arguments;
      int index000 = 0;
      int length000 = vector000->length();

      while (index000 < length000) {
        arg = (vector000->theArray)[index000];
        index000 = index000 + 1;
        { Surrogate* testValue000 = safePrimaryType(arg);

          if (subtypeOfP(testValue000, SGT_INFERENCE_CACHES_LOGIC_DESCRIPTION)) {
            { Object* arg000 = arg;
              Description* arg = ((Description*)(arg000));

              pointstometaobjectP = TRUE;
            }
          }
          else if (subtypeOfP(testValue000, SGT_INFERENCE_CACHES_STELLA_LITERAL_WRAPPER)) {
            { Object* arg001 = arg;
              LiteralWrapper* arg = ((LiteralWrapper*)(arg001));

            }
          }
          else {
            return (FALSE);
          }
        }
      }
    }
    return (pointstometaobjectP);
  }
}

void handleOutOfDateInferenceCache(Keyword* assertorretract, Proposition* proposition) {
  if (worldStateP(oCONTEXTo.get())) {
    if ((!((boolean)(lookupInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_META)))) ||
        (((boolean)(oQUERYITERATORo.get())) ||
         (descriptionModeP() ||
          oINVISIBLEASSERTIONpo.get()))) {
      return;
    }
    { NamedDescription* description = (((boolean)(proposition)) ? getDescription(((Surrogate*)(proposition->operatoR))) : ((NamedDescription*)(NULL)));

      if (((boolean)(description)) &&
          testPropertyP(description, SGT_INFERENCE_CACHES_PL_KERNEL_KB_HIDDEN_RELATION)) {
        return;
      }
    }
    if (metaPropositionP(proposition) &&
        ((assertorretract == KWD_INFERENCE_CACHES_RETRACT) &&
         (!((BooleanWrapper*)(dynamicSlotValue(lookupInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_META)->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINEDp, FALSE_WRAPPER)))->wrapperValue))) {
      if (((boolean)(oTRACED_KEYWORDSo)) &&
          oTRACED_KEYWORDSo->membP(KWD_INFERENCE_CACHES_PERFORMANCE_CLUES)) {
        std::cout << "DELETING META CACHE" << std::endl;
      }
      destroyInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_META);
    }
    else if (oJUST_IN_TIME_FORWARD_INFERENCEpo) {
      if (((boolean)(((World*)(dynamicSlotValue(oCONTEXTo.get()->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE, NULL)))))) {
        if (((boolean)(oTRACED_KEYWORDSo)) &&
            oTRACED_KEYWORDSo->membP(KWD_INFERENCE_CACHES_PERFORMANCE_CLUES)) {
          std::cout << "DELETING JIT CACHE" << std::endl;
        }
        destroyInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_JUST_IN_TIME);
      }
    }
    else {
      { World* cache = lookupInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_TMS);

        if (((boolean)(cache)) &&
            (!((BooleanWrapper*)(dynamicSlotValue(cache->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINEDp, FALSE_WRAPPER)))->wrapperValue)) {
          destroyInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_TMS);
        }
      }
    }
    return;
  }
  { World* world = ((World*)(oCONTEXTo.get()));

    if (((BooleanWrapper*)(dynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_MONOTONICp, FALSE_WRAPPER)))->wrapperValue &&
        (assertorretract == KWD_INFERENCE_CACHES_RETRACT)) {
      *(STANDARD_WARNING->nativeStream) << "Warning: " << "Retraction occurred in world " << "`" << world << "'" << " which is marked monotonic." << std::endl;
    }
    if (!((BooleanWrapper*)(dynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINEDp, FALSE_WRAPPER)))->wrapperValue) {
      setDynamicSlotValue(world->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
  }
}

World* bestInferenceCache(Context* context) {
  if (oJUST_IN_TIME_FORWARD_INFERENCEpo) {
    return (getInferenceCache(context, KWD_INFERENCE_CACHES_JUST_IN_TIME));
  }
  { World* cache = lookupInferenceCache(context, KWD_INFERENCE_CACHES_TMS);

    if (((boolean)(cache))) {
      return (cache);
    }
    else {
      return (lookupInferenceCache(context, KWD_INFERENCE_CACHES_META));
    }
  }
}

Context* getQueryContext() {
  if (!worldStateP(oCONTEXTo.get())) {
    return (oCONTEXTo.get());
  }
  { World* cache = bestInferenceCache(oCONTEXTo.get());

    return ((((boolean)(cache)) ? cache : oCONTEXTo.get()));
  }
}

Context* getPropertyTestContext() {
  { World* constraintpropagationworld = lookupConstraintPropagationWorld(oCONTEXTo.get());

    if (!((boolean)(constraintpropagationworld))) {
      return (oCONTEXTo.get());
    }
    else {
      return (constraintpropagationworld);
    }
  }
}

Context* getInferableTestContext() {
  if (!worldStateP(oCONTEXTo.get())) {
    return (oCONTEXTo.get());
  }
  else {
    { World* temp000 = lookupInferenceCache(oCONTEXTo.get(), KWD_INFERENCE_CACHES_META);

      { Context* value000 = (((boolean)(temp000)) ? temp000 : oCONTEXTo.get());

        return (value000);
      }
    }
  }
}

Context* getTopLevelUpdateContext() {
  return (getWorldState(oCONTEXTo.get()));
}

World* lookupConstraintPropagationWorld(Context* self) {
  if (worldStateP(self)) {
    if (oJUST_IN_TIME_FORWARD_INFERENCEpo) {
      if (((boolean)(oQUERYITERATORo.get()))) {
        return (getInferenceCache(self, KWD_INFERENCE_CACHES_JUST_IN_TIME));
      }
      else {
        return (lookupInferenceCache(self, KWD_INFERENCE_CACHES_JUST_IN_TIME));
      }
    }
    else {
      return (lookupInferenceCache(self, KWD_INFERENCE_CACHES_TMS));
    }
  }
  else {
    return (((World*)(self)));
  }
}

void inheritCachedInferences(World* fromworld, World* toworld) {
  toworld = toworld;
  if (!((boolean)(fromworld))) {
    return;
  }
}

void inheritTruthValueUpdate(Proposition* proposition, World* toworld) {
}

void inheritVariableValueUpdate(Skolem* skolem, World* toworld) {
}

void callPropagateConstraints(Context* context) {
  // Trigger constraint propagation over all propositions
  // in the module or world `context'.
  { Context* worldstate = getWorldState(context);
    World* propagationworld = getInferenceCache(worldstate, KWD_INFERENCE_CACHES_META);

    if (((BooleanWrapper*)(dynamicSlotValue(propagationworld->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, FALSE_WRAPPER)))->wrapperValue) {
      return;
    }
    try {
      { Proposition* p = NULL;
        Iterator* iter000 = allPropositions(((Module*)(worldstate)), TRUE);

        while (iter000->nextP()) {
          p = ((Proposition*)(iter000->value));
          reactToKbUpdate(propagationworld, p);
        }
      }
    }
    catch (Clash& _e) {
      Clash* e = &_e;

      *(STANDARD_ERROR->nativeStream) << exceptionMessage(e) << std::endl;
      *(STANDARD_ERROR->nativeStream) << "One or more facts need to be retracted to eliminate" << std::endl << "   the inconsistency." << std::endl;
    }
    setDynamicSlotValue(propagationworld->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
  }
}

void propagateConstraints(Cons* name) {
  // Trigger constraint propagation over all propositions of module `name'.
  // If no `name' is supplied, the current module will be used.  This also
  // enables incremental constraint propagation for future monotonic updates to
  // the module.  Once a non-monotonic update is performed, i.e., a retraction
  // or clipping of a function value, all cached inferences will be discarded
  // and constraint propagation will be turned off until this function is
  // called again.
  processCheckTypesAgenda();
  { Cons* optionalname = name;
    Module* module = NULL;

    if (((boolean)(optionalname->value))) {
      module = coerceToModule(optionalname->value, TRUE);
    }
    if (((boolean)(module))) {
      callPropagateConstraints(module);
    }
    else {
      callPropagateConstraints(oCONTEXTo.get());
    }
  }
}

void propagateConstraintsEvaluatorWrapper(Cons* arguments) {
  propagateConstraints(arguments);
}

void repropagateConstraints(Cons* name) {
  // Force non-incremental constraint propagation over all propositions of
  // module `name'.  If no `name' is supplied, the current module will be used.
  // This also enables incremental constraint propagation for future monotonic
  // updates to the module similar to `propagate-constraints'.
  processCheckTypesAgenda();
  { Cons* optionalname = name;
    Module* context = coerceToModule(optionalname->value, TRUE);

    if (!((boolean)(context))) {
      context = ((Module*)(oCONTEXTo.get()));
    }
    destroyInferenceCaches(context);
    callPropagateConstraints(context);
  }
}

void repropagateConstraintsEvaluatorWrapper(Cons* arguments) {
  repropagateConstraints(arguments);
}

void callRunForwardRules(Module* module, boolean forceP) {
  // Run forward inference rules in module 'module'. If 'module'
  // is NULL, the current module will be used.  If forward inferencing is already
  // up-to-date in the designated module, no additional inferencing will occur,
  // unless 'force?' is set to TRUE, in which case all forward rules are run or rerun.
  finalizeObjects();
  processCheckTypesAgenda();
  if (forceP) {
    destroyInferenceCaches(module);
  }
  { Context* worldstate = getWorldState(module);
    World* propagationworld = getInferenceCache(worldstate, KWD_INFERENCE_CACHES_META);

    if (((BooleanWrapper*)(dynamicSlotValue(propagationworld->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, FALSE_WRAPPER)))->wrapperValue) {
      return;
    }
    { Proposition* p = NULL;
      Iterator* iter000 = allPropositions(((Module*)(worldstate)), TRUE);

      while (iter000->nextP()) {
        p = ((Proposition*)(iter000->value));
        reactToKbUpdate(propagationworld, p);
      }
    }
    setDynamicSlotValue(propagationworld->dynamicSlots, SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
  }
}

void runForwardRules(Object* moduleref, Cons* force) {
  // Run forward inference rules in module 'moduleRef'. If 'moduleRef' is NULL, the
  // current module will be used.  If forward inferencing is already up-to-date
  // in the designated module, no additional inferencing will occur, unless the
  // optional keyword `:force' is included, in which case all
  // forward rules are run or rerun.
  // 
  // Calling `run-forward-rules' temporarily puts the module into a mode where
  // future assertional (monotonic) updates will trigger additional forward
  // inference.  Once a non-monotonic update is performed, i.e., a retraction
  // or clipping of relation value, all cached forward inferences will be discarded
  // and forward inferencing will be disabled until this function is
  // called again.
  { Module* module = coerceToModule(moduleref, TRUE);
    Cons* optionalarguments = force;
    boolean forceP = ((Keyword*)(optionalarguments->value)) == KWD_INFERENCE_CACHES_FORCE;

    if (!((boolean)(module))) {
      module = oMODULEo.get();
    }
    try {
      callRunForwardRules(module, forceP);
    }
    catch (Clash& _e) {
      Clash* e = &_e;

      *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
      std::cout << "   One or more facts need to be retracted to eliminate" << " the inconsistency." << std::endl;
    }
  }
}

void runForwardRulesEvaluatorWrapper(Cons* arguments) {
  runForwardRules(arguments->value, arguments->rest);
}

void startupInferenceCaches() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, getStellaModule("/LOGIC", oSTARTUP_TIME_PHASEo > 1));
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (currentStartupTimePhaseP(2)) {
      SGT_INFERENCE_CACHES_STELLA_WORLD = ((Surrogate*)(internRigidSymbolWrtModule("WORLD", getStellaModule("/STELLA", TRUE), 1)));
      SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF = ((Symbol*)(internRigidSymbolWrtModule("INFERENCE-CACHE-OF", NULL, 0)));
      SYM_INFERENCE_CACHES_LOGIC_MONOTONICp = ((Symbol*)(internRigidSymbolWrtModule("MONOTONIC?", NULL, 0)));
      KWD_INFERENCE_CACHES_JUST_IN_TIME = ((Keyword*)(internRigidSymbolWrtModule("JUST-IN-TIME", NULL, 2)));
      SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE = ((Symbol*)(internRigidSymbolWrtModule("JUST-IN-TIME-INFERENCE-CACHE", NULL, 0)));
      KWD_INFERENCE_CACHES_META = ((Keyword*)(internRigidSymbolWrtModule("META", NULL, 2)));
      SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE = ((Symbol*)(internRigidSymbolWrtModule("META-INFERENCE-CACHE", NULL, 0)));
      KWD_INFERENCE_CACHES_TMS = ((Keyword*)(internRigidSymbolWrtModule("TMS", NULL, 2)));
      SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE = ((Symbol*)(internRigidSymbolWrtModule("TRUTH-MAINTAINED-INFERENCE-CACHE", NULL, 0)));
      SYM_INFERENCE_CACHES_LOGIC_EVALUATION_STATE_TABLE = ((Symbol*)(internRigidSymbolWrtModule("EVALUATION-STATE-TABLE", NULL, 0)));
      SYM_INFERENCE_CACHES_LOGIC_EVALUATION_QUEUE = ((Symbol*)(internRigidSymbolWrtModule("EVALUATION-QUEUE", NULL, 0)));
      SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp = ((Symbol*)(internRigidSymbolWrtModule("UP-TO-DATE-INFERENCES?", NULL, 0)));
      SGT_INFERENCE_CACHES_LOGIC_DESCRIPTION = ((Surrogate*)(internRigidSymbolWrtModule("DESCRIPTION", NULL, 1)));
      SGT_INFERENCE_CACHES_STELLA_LITERAL_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("LITERAL-WRAPPER", getStellaModule("/STELLA", TRUE), 1)));
      SGT_INFERENCE_CACHES_PL_KERNEL_KB_HIDDEN_RELATION = ((Surrogate*)(internRigidSymbolWrtModule("HIDDEN-RELATION", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
      KWD_INFERENCE_CACHES_RETRACT = ((Keyword*)(internRigidSymbolWrtModule("RETRACT", NULL, 2)));
      SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINEDp = ((Symbol*)(internRigidSymbolWrtModule("TRUTH-MAINTAINED?", NULL, 0)));
      KWD_INFERENCE_CACHES_PERFORMANCE_CLUES = ((Keyword*)(internRigidSymbolWrtModule("PERFORMANCE-CLUES", NULL, 2)));
      KWD_INFERENCE_CACHES_FORCE = ((Keyword*)(internRigidSymbolWrtModule("FORCE", NULL, 2)));
      SYM_INFERENCE_CACHES_LOGIC_STARTUP_INFERENCE_CACHES = ((Symbol*)(internRigidSymbolWrtModule("STARTUP-INFERENCE-CACHES", NULL, 0)));
      SYM_INFERENCE_CACHES_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(internRigidSymbolWrtModule("METHOD-STARTUP-CLASSNAME", getStellaModule("/STELLA", TRUE), 0)));
    }
    if (currentStartupTimePhaseP(4)) {
      oINVISIBLEASSERTIONpo.set(FALSE);
    }
    if (currentStartupTimePhaseP(6)) {
      finalizeClasses();
    }
    if (currentStartupTimePhaseP(7)) {
      defineFunctionObject("WORLD-STATE?", "(DEFUN (WORLD-STATE? BOOLEAN) ((SELF CONTEXT)) :PUBLIC? TRUE)", ((cpp_function_code)(&worldStateP)), NULL);
      defineFunctionObject("GET-WORLD-STATE", "(DEFUN (GET-WORLD-STATE CONTEXT) ((SELF CONTEXT)) :PUBLIC? TRUE)", ((cpp_function_code)(&getWorldState)), NULL);
      defineExternalSlotFromStringifiedSource("(DEFSLOT CONTEXT META-INFERENCE-CACHE :TYPE WORLD :ALLOCATION :DYNAMIC)");
      defineExternalSlotFromStringifiedSource("(DEFSLOT CONTEXT TRUTH-MAINTAINED-INFERENCE-CACHE :TYPE WORLD :ALLOCATION :DYNAMIC)");
      defineExternalSlotFromStringifiedSource("(DEFSLOT CONTEXT JUST-IN-TIME-INFERENCE-CACHE :TYPE WORLD :ALLOCATION :DYNAMIC)");
      defineExternalSlotFromStringifiedSource("(DEFSLOT WORLD INFERENCE-CACHE-OF :TYPE CONTEXT :PUBLIC? TRUE :ALLOCATION :DYNAMIC)");
      defineFunctionObject("LOOKUP-INFERENCE-CACHE", "(DEFUN (LOOKUP-INFERENCE-CACHE WORLD) ((CONTEXT CONTEXT) (KIND KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&lookupInferenceCache)), NULL);
      defineFunctionObject("INITIALIZE-INFERENCE-WORLD", "(DEFUN INITIALIZE-INFERENCE-WORLD ((WORLD WORLD)))", ((cpp_function_code)(&initializeInferenceWorld)), NULL);
      defineFunctionObject("CREATE-INFERENCE-CACHE", "(DEFUN (CREATE-INFERENCE-CACHE WORLD) ((CONTEXT CONTEXT) (KIND KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&createInferenceCache)), NULL);
      defineFunctionObject("GET-INFERENCE-CACHE", "(DEFUN (GET-INFERENCE-CACHE WORLD) ((CONTEXT CONTEXT) (KIND KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&getInferenceCache)), NULL);
      defineFunctionObject("DESTROY-INFERENCE-CACHE", "(DEFUN DESTROY-INFERENCE-CACHE ((CONTEXT CONTEXT) (KIND KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&destroyInferenceCache)), NULL);
      defineFunctionObject("UNLINK-INFERENCE-CACHE", "(DEFUN UNLINK-INFERENCE-CACHE ((SELF WORLD)))", ((cpp_function_code)(&unlinkInferenceCache)), NULL);
      defineFunctionObject("DESTROY-INFERENCE-CACHES", "(DEFUN DESTROY-INFERENCE-CACHES ((CONTEXT CONTEXT)) :PUBLIC? TRUE)", ((cpp_function_code)(&destroyInferenceCaches)), NULL);
      defineFunctionObject("META-PROPOSITION?", "(DEFUN (META-PROPOSITION? BOOLEAN) ((PROPOSITION PROPOSITION)) :PUBLIC? TRUE)", ((cpp_function_code)(&metaPropositionP)), NULL);
      defineFunctionObject("HANDLE-OUT-OF-DATE-INFERENCE-CACHE", "(DEFUN HANDLE-OUT-OF-DATE-INFERENCE-CACHE ((ASSERTORRETRACT KEYWORD) (PROPOSITION PROPOSITION)))", ((cpp_function_code)(&handleOutOfDateInferenceCache)), NULL);
      defineFunctionObject("BEST-INFERENCE-CACHE", "(DEFUN (BEST-INFERENCE-CACHE WORLD) ((CONTEXT CONTEXT)))", ((cpp_function_code)(&bestInferenceCache)), NULL);
      defineFunctionObject("GET-QUERY-CONTEXT", "(DEFUN (GET-QUERY-CONTEXT CONTEXT) () :PUBLIC? TRUE)", ((cpp_function_code)(&getQueryContext)), NULL);
      defineFunctionObject("GET-PROPERTY-TEST-CONTEXT", "(DEFUN (GET-PROPERTY-TEST-CONTEXT CONTEXT) ())", ((cpp_function_code)(&getPropertyTestContext)), NULL);
      defineFunctionObject("GET-INFERABLE-TEST-CONTEXT", "(DEFUN (GET-INFERABLE-TEST-CONTEXT CONTEXT) ())", ((cpp_function_code)(&getInferableTestContext)), NULL);
      defineFunctionObject("GET-TOP-LEVEL-UPDATE-CONTEXT", "(DEFUN (GET-TOP-LEVEL-UPDATE-CONTEXT CONTEXT) ())", ((cpp_function_code)(&getTopLevelUpdateContext)), NULL);
      defineFunctionObject("LOOKUP-CONSTRAINT-PROPAGATION-WORLD", "(DEFUN (LOOKUP-CONSTRAINT-PROPAGATION-WORLD WORLD) ((SELF CONTEXT)))", ((cpp_function_code)(&lookupConstraintPropagationWorld)), NULL);
      defineFunctionObject("INHERIT-CACHED-INFERENCES", "(DEFUN INHERIT-CACHED-INFERENCES ((FROMWORLD WORLD) (TOWORLD WORLD)))", ((cpp_function_code)(&inheritCachedInferences)), NULL);
      defineFunctionObject("INHERIT-TRUTH-VALUE-UPDATE", "(DEFUN INHERIT-TRUTH-VALUE-UPDATE ((PROPOSITION PROPOSITION) (TOWORLD WORLD)))", ((cpp_function_code)(&inheritTruthValueUpdate)), NULL);
      defineFunctionObject("INHERIT-VARIABLE-VALUE-UPDATE", "(DEFUN INHERIT-VARIABLE-VALUE-UPDATE ((SKOLEM SKOLEM) (TOWORLD WORLD)))", ((cpp_function_code)(&inheritVariableValueUpdate)), NULL);
      defineFunctionObject("CALL-PROPAGATE-CONSTRAINTS", "(DEFUN CALL-PROPAGATE-CONSTRAINTS ((CONTEXT CONTEXT)) :DOCUMENTATION \"Trigger constraint propagation over all propositions\nin the module or world `context'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&callPropagateConstraints)), NULL);
      defineFunctionObject("PROPAGATE-CONSTRAINTS", "(DEFUN PROPAGATE-CONSTRAINTS (|&REST| (NAME NAME)) :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Trigger constraint propagation over all propositions of module `name'.\nIf no `name' is supplied, the current module will be used.  This also\nenables incremental constraint propagation for future monotonic updates to\nthe module.  Once a non-monotonic update is performed, i.e., a retraction\nor clipping of a function value, all cached inferences will be discarded\nand constraint propagation will be turned off until this function is\ncalled again.\")", ((cpp_function_code)(&propagateConstraints)), ((cpp_function_code)(&propagateConstraintsEvaluatorWrapper)));
      defineFunctionObject("REPROPAGATE-CONSTRAINTS", "(DEFUN REPROPAGATE-CONSTRAINTS (|&REST| (NAME NAME)) :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Force non-incremental constraint propagation over all propositions of\nmodule `name'.  If no `name' is supplied, the current module will be used.\nThis also enables incremental constraint propagation for future monotonic\nupdates to the module similar to `propagate-constraints'.\")", ((cpp_function_code)(&repropagateConstraints)), ((cpp_function_code)(&repropagateConstraintsEvaluatorWrapper)));
      defineFunctionObject("CALL-RUN-FORWARD-RULES", "(DEFUN CALL-RUN-FORWARD-RULES ((MODULE MODULE) (FORCE? BOOLEAN)) :DOCUMENTATION \"Run forward inference rules in module 'module'. If 'module'\nis NULL, the current module will be used.  If forward inferencing is already\nup-to-date in the designated module, no additional inferencing will occur,\nunless 'force?' is set to TRUE, in which case all forward rules are run or rerun.\" :PUBLIC? TRUE)", ((cpp_function_code)(&callRunForwardRules)), NULL);
      defineFunctionObject("RUN-FORWARD-RULES", "(DEFUN RUN-FORWARD-RULES ((MODULEREF NAME) |&REST| (FORCE KEYWORD)) :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Run forward inference rules in module 'moduleRef'. If 'moduleRef' is NULL, the\ncurrent module will be used.  If forward inferencing is already up-to-date\nin the designated module, no additional inferencing will occur, unless the\noptional keyword `:force' is included, in which case all\nforward rules are run or rerun.\n\nCalling `run-forward-rules' temporarily puts the module into a mode where\nfuture assertional (monotonic) updates will trigger additional forward\ninference.  Once a non-monotonic update is performed, i.e., a retraction\nor clipping of relation value, all cached forward inferences will be discarded\nand forward inferencing will be disabled until this function is\ncalled again.\")", ((cpp_function_code)(&runForwardRules)), ((cpp_function_code)(&runForwardRulesEvaluatorWrapper)));
      defineFunctionObject("STARTUP-INFERENCE-CACHES", "(DEFUN STARTUP-INFERENCE-CACHES () :PUBLIC? TRUE)", ((cpp_function_code)(&startupInferenceCaches)), NULL);
      { MethodSlot* function = lookupFunction(SYM_INFERENCE_CACHES_LOGIC_STARTUP_INFERENCE_CACHES);

        setDynamicSlotValue(function->dynamicSlots, SYM_INFERENCE_CACHES_STELLA_METHOD_STARTUP_CLASSNAME, wrapString("_Startup-Inference-Caches"), NULL_STRING_WRAPPER);
      }
    }
    if (currentStartupTimePhaseP(8)) {
      finalizeSlots();
      cleanupUnfinalizedClasses();
    }
    if (currentStartupTimePhaseP(9)) {
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *INVISIBLEASSERTION?* BOOLEAN FALSE :DOCUMENTATION \"Used to signal an assertion that doesn't\nblow away inference caches.\")");
    }
  }
}

Surrogate* SGT_INFERENCE_CACHES_STELLA_WORLD = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_INFERENCE_CACHE_OF = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_MONOTONICp = NULL;

Keyword* KWD_INFERENCE_CACHES_JUST_IN_TIME = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_JUST_IN_TIME_INFERENCE_CACHE = NULL;

Keyword* KWD_INFERENCE_CACHES_META = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_META_INFERENCE_CACHE = NULL;

Keyword* KWD_INFERENCE_CACHES_TMS = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_EVALUATION_STATE_TABLE = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_EVALUATION_QUEUE = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_UP_TO_DATE_INFERENCESp = NULL;

Surrogate* SGT_INFERENCE_CACHES_LOGIC_DESCRIPTION = NULL;

Surrogate* SGT_INFERENCE_CACHES_STELLA_LITERAL_WRAPPER = NULL;

Surrogate* SGT_INFERENCE_CACHES_PL_KERNEL_KB_HIDDEN_RELATION = NULL;

Keyword* KWD_INFERENCE_CACHES_RETRACT = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_TRUTH_MAINTAINEDp = NULL;

Keyword* KWD_INFERENCE_CACHES_PERFORMANCE_CLUES = NULL;

Keyword* KWD_INFERENCE_CACHES_FORCE = NULL;

Symbol* SYM_INFERENCE_CACHES_LOGIC_STARTUP_INFERENCE_CACHES = NULL;

Symbol* SYM_INFERENCE_CACHES_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace logic
